 ! Fermi-Pasta-Ulam problem
 
 module comm
   implicit none
   
   real(kind=8), parameter::pi=acos(-1.d0)
   integer(kind=8)::N ! number of particles
   real(kind=8), parameter::r=3.d0, alpha=0.1d0 ! potential parameters
   
   real(kind=8)::L ! time interval
   integer(kind=8)::S ! number of steps on the time axis
   real(kind=8)::h ! step on the time axis
 
 contains
   
   ! subroutine for force computation
   subroutine forces(q, f)
     implicit none
     
     real(kind=8), dimension(0:N-1), intent(in)::q
     real(kind=8), dimension(0:N-1), intent(out)::f
     integer(kind=8)::i
     
     f(0)=-2.d0*q(0)+q(N-1)+q(1)+alpha*((q(N-1)-q(0))**(r-1.d0)-(q(0)-q(1))**(r-1.d0))
     do i=1, N-2
       f(i)=-2.d0*q(i)+q(i-1)+q(i+1)+alpha*((q(i-1)-q(i))**(r-1.d0)-(q(i)-q(i+1))**(r-1.d0))
     enddo
     f(N-1)=-2.d0*q(N-1)+q(N-2)+q(0)+alpha*((q(N-2)-q(N-1))**(r-1.d0)-(q(N-1)-q(0))**(r-1.d0))
   endsubroutine
   
   
   ! velocity Verlet
   subroutine verlet(q, p, f)
     implicit none
   
     real(kind=8), dimension(0:N-1)::q, p, f
     integer(kind=8)::i
     
     do i=0, N-1
       p(i)=p(i)+0.5d0*h*f(i)
       q(i)=q(i)+h*p(i)
     enddo
     
     call forces(q, f)
     
     do i=0, N-1
       p(i)=p(i)+0.5d0*h*f(i)
     enddo
   endsubroutine
   
   ! this function returns an array containing the N energies of the normal modes
   subroutine compute_nmodes_energies_fftw(q, p, energies, mode)
     implicit none
     inCLUDE 'fftw3.f'
     
     real(kind=8), dimension(0:), intent(in) ::q,p
     LOGICAL,intent(in)                      ::mode
     real(kind=8), dimension(0:), intent(out)::energies
     
     COMPLEX(kind=8), dimension(0:N-1)::qc1,qc2,pc1,pc2
     real(kind=8), dimension(0:N-1)::aq, ap
     integer(kind=8)::M,i,j,plan

     M=size(energies)
     
     ! normal modes computation
     qc1=dcmplx(q)
     call dfftw_plan_dft_1d(plan,N,qc1,qc2,FFTW_forward,fftw_estimate)
     call dfftw_execute(plan)
     call dfftw_destroy_plan(plan)

     pc1=dcmplx(p)
     call dfftw_plan_dft_1d(plan,N,pc1,pc2,FFTW_forward,fftw_estimate)
     call dfftw_execute(plan)
     call dfftw_destroy_plan(plan)
     
     aq=(2.d0/dble(N))*dimag(qc2)
     ap=(2.d0/dble(N))*dimag(pc2)

     if (mode.eqv..false.) then
       do i=0, M-1
         energies(i)=(0.5*ap(i)**2+2*sin(i*pi/N)**2*aq(i)**2)
       enddo
     else
       do i=0, M-1
         energies(i)=0
         do j=0, N-1
           energies(i)=energies(i)+0.5d0*sin(2*pi*i*j/dble(N))**2*(ap(i)**2+(2*sin(pi*i/dble(N))*aq(i))**2)
         enddo
       enddo
     endif
   endsubroutine

   ! average energies
   subroutine average_energies(current_energies,current_time,avg_energies)
     implicit none

     real(kind=8),intent(in)::current_energies(:),current_time
     real(kind=8),intent(out)::avg_energies(:)
     
     real(kind=8),save,allocatable::sum_energies(:),prev_energies(:),prev_time
     logical,save::first=.true.

     integer::M
     
     if (first.eqv..true.) then
       first=.false.
       M=size(current_energies)
       allocate(sum_energies(M),prev_energies(M))
       prev_energies=0
       prev_time=0
       sum_energies=0
     endif
     
     sum_energies=sum_energies+(prev_energies+current_energies)*(current_time-prev_time)*0.5d0
     avg_energies=sum_energies/current_time

     prev_energies=current_energies
     prev_time=current_time

   endsubroutine

   ! total energy
   function tot_energy (q,p)
     implicit none
     
     real(kind=8), dimension(0:N-1)::q,p
     real(kind=8)::tot_energy
     integer(kind=8)::i
     
     tot_energy=0.5d0*p(N-1)**2+0.5d0*(q(0)-q(N-1))**2+alpha/r*(q(N-1)-q(0))**r
     do i=0, N-2
       tot_energy=tot_energy+0.5d0*p(i)**2+0.5d0*(q(i)-q(i+1))**2+alpha/r*(q(i)-q(i+1))**r
     enddo
     
     return
   endfunction
   
   ! kinetic energy
   function kin_energy (p)
     implicit none
     
     real(kind=8), dimension(0:N-1)::p
     real(kind=8)::kin_energy
     integer(kind=8)::i
     
     kin_energy=0.d0
     do i=0, N-1
       kin_energy=kin_energy+0.5d0*p(i)**2
     enddo
     
     return
   endfunction
   
   ! total momentum
   function momentum (p)
     implicit none
     
     real(kind=8), dimension(0:N-1)::p
     real(kind=8)::momentum
     integer(kind=8)::i
     
     momentum=0.d0
     do i=0, N-1
       momentum=momentum+p(i)
     enddo
     
     return
   endfunction
   
   function spectral_entropy(energies)
     implicit none
     
     real(kind=8), dimension(0:N/2-1), intent(in)::energies
     real(kind=8)::spectral_entropy
     
     real(kind=8), dimension(0:N/2-1)::p
     integer::i
     
     p=energies/sum(energies)
     
     spectral_entropy=0.d0
     do i=0, N/2-1
       if (p(i) /= 0) then
         spectral_entropy=spectral_entropy-p(i)*log(p(i))
       endif
     enddo
     
     return
   endfunction
   
   function eta(entropy, zero_entropy)
     real(kind=8), intent(in)::entropy, zero_entropy
     real(kind=8)::eta
     
     eta=(log(dble(N))-entropy)/(log(dble(N))-zero_entropy)
     
     return
   endfunction
   
 endmodule
 
 program main
   use comm
   implicit none
   
   ! main program variables
   real(kind=8), dimension(:), allocatable:: x, v, f, t ! grids
   real(kind=8), dimension(:), allocatable:: nmodes_energies,average_nmodes_energies
   real(kind=8)::zero_entropy
   integer::nmodes_out ! number of normal modes you want to print
   integer::i, j ! indexes
   
   ! read grid parameters
   print *, 'Reading grid parameters...'
   open(unit=1, file='input/input_data.dat')
   read(1,*)
   read (1,*) N, L, S
   read(1,*)
   read(1,*)
   read(1,*)
   read(1,*)nmodes_out
   close(unit=1)
   
   ! create grid
   print *, 'Creating grid...'
   h=L/dble(S)
   allocate (x(0:N-1), v(0:N-1), f(0:N-1), t(0:S-1), nmodes_energies(0:N/2-1),average_nmodes_energies(0:N/2-1))
   do i=0, S-1
     t(i)=h*i
   enddo
   
   ! initial conditions
   print *, 'Reading initial conditions...'
   open(unit=1, file='input/ic.dat')
   do i=0, N-1
     read(1,*) j, x(i), v(i)
   enddo
   close(unit=1)
   
   call compute_nmodes_energies_fftw(x,v,nmodes_energies,.true.)
   zero_entropy=spectral_entropy(nmodes_energies)
   
   print *, 'Integrating equations and producing output...'
   ! open(unit=1, file='output/coordinates.dat')
   ! open(unit=2, file='output/velocities.dat')
   open(unit=3, file='output/first_integrals.dat')
   open(unit=4, file='output/nmodes_energies.dat')
   open(unit=5, file='output/averaged_energies.dat')
   open(unit=6, file='output/spectral_entropy.dat')
   ! write(1,*) t(0), x
   ! write(2,*) t(0), v
   write(3,*) t(0), kin_energy(v), tot_energy(x, v), momentum (v)
   write(4,*) t(0), nmodes_energies(0:nmodes_out), &
     sum(nmodes_energies(nmodes_out:N/2-1)), sum(nmodes_energies)
   write(5,*) t(0), 0
   write(6,*) t(0), zero_entropy, 1
   
   call forces(x, f)
   do i=1, S-1
     call verlet(x, v, f)
     call compute_nmodes_energies_fftw(x,v,nmodes_energies,.true.)
     call average_energies(nmodes_energies,t(i),average_nmodes_energies)
   
     ! write(1,*) t(i), x
     ! write(2,*) t(i), v
     write(3,*) t(i), kin_energy(v), tot_energy(x, v), momentum (v)
     write(4,*) t(i), nmodes_energies(0:nmodes_out), &
       sum(nmodes_energies(nmodes_out:N/2-1)), sum(nmodes_energies)
     write(5,*) t(i), average_nmodes_energies(0:nmodes_out)
     write(6,*) t(i), spectral_entropy(nmodes_energies), &
       eta(spectral_entropy(nmodes_energies), zero_entropy)
   enddo
   
   ! close(unit=1)
   ! close(unit=2)
   close(unit=3)
   close(unit=4)
   close(unit=5)
   close(unit=6)
   
endprogram
